{*******************************************************}
{                                                       }
{       Turbo Vision Unit                               }
{                                                       }
{*******************************************************}

unit Console;

{$O+,X+,I-,S-}

interface

uses Objects,Drivers,Views, WinDos;

Const
	TextAttr: Word = 7;

Procedure ClrScr;
Procedure ClrEOL;
Procedure GotoXY(x, y: Integer);
Function WhereX: Integer;
Function WhereY: Integer;

Type
	{ TTextDevice }
	PTextDevice = ^TTextDevice;
	TTextDevice = object(TScroller)
		function StrRead(s: PChar): Byte; virtual;
		procedure StrWrite(s: PChar; Count: Word); virtual;
	end;

	{ TTerminal }
	PTerminalBuffer = ^TTerminalBuffer;
	TTerminalBuffer = array[0..32757] of Word;

	PTerminal = ^TTerminal;
	TTerminal = object(TTextDevice)
		BufSize: Word;
		KeyBufLen: Integer;
		Buffer: PTerminalBuffer;
		KeyBuf: Array[0..31]Of Char;
		PrevCC: PTerminal;
		XRes, YRes: Integer;
		constructor Init(var Bounds:TRect;AHSB,AVSB:PScrollBar;AXRes,AYRes:Word);
		destructor Done; virtual;
		procedure Draw; virtual;
		Procedure HandleEvent(Var Event: TEvent); Virtual;
		function StrRead(s: PChar): Byte; virtual;
		procedure StrWrite(s: PChar; Count: Word); virtual;
	end;

Const
	CC: PTerminal = Nil;

procedure AssignDevice(var T: Text; Screen: PTextDevice);

implementation

Procedure ClrEOL;
Begin
	FillWord(
		CC^.Buffer^[CC^.Cursor.Y*CC^.XRes+CC^.Cursor.X],
		CC^.XRes-CC^.Cursor.X,TextAttr Shl 8+32);
	CC^.DrawView
End;
Procedure ClrScr;Begin FillWord(CC^.Buffer,CC^.BufSize Div 2,TextAttr Shl 8+32); GotoXY(1,1)End;
Procedure GotoXY;Begin CC^.SetCursor(x-1, y-1)End;
Function WhereX; Begin WhereX := CC^.Cursor.X+1 End;
Function WhereY; Begin WhereY := CC^.Cursor.Y+1 End;

{ TTextDevice }
function TTextDevice.StrRead; Begin StrRead := 0 End;
procedure TTextDevice.StrWrite; Begin End;

{ TTerminal }
constructor TTerminal.Init;
begin
	TTextDevice.Init(Bounds, AHSB, AVSB);
	GrowMode := gfGrowHiX + gfGrowHiY;
	Asm
		les di, Self
		mov ax, AXRes; mov es:[di].TTerminal.XRes, ax
		mov bx, AYRes; mov es:[di].TTerminal.YRes, bx
		mul bx
		shl ax, 1
		mov es:[di].TTerminal.BufSize, ax
	End;
	GetMem(Buffer, BufSize);
	FillWord(Buffer^, AXRes*AYRes, TextAttr Shl 8+32);

	PrevCC := CC;
	CC := @Self;

	SetLimit(AXRes, AYRes);
	KeyBufLen := 0;
	ScrollTo(0, 0);
end;

destructor TTerminal.Done;
begin
	CC := PrevCC;
	FreeMem(Buffer, BufSize);
	TTextDevice.Done;
end;

procedure TTerminal.Draw;
var
	I: Integer;
	B: TDrawBuffer;
Begin
	For I := 0 To Size.Y-1 Do
	Begin
		MoveChar(B, ' ', 7, Size.X);
		If Delta.Y+I >= Limit.Y-1
			Then WriteLine(0,I, Size.X-Delta.X,1, B)
			Else WriteLine(0,I, Size.X-Delta.X,1, Buffer^[(I+Delta.Y)*XRes+Delta.X])
	End;
	ShowCursor;
	NormalCursor
End;

function TTerminal.StrRead(s: PChar): Byte;
Var Event: TEvent;
Begin
	StrRead := KeyBufLen;
	Move(KeyBuf, s^, KeyBufLen);
	KeyBufLen := 0
End;

Procedure TTerminal.HandleEvent;
Begin
	Inherited HandleEvent(Event);
	If Event.What = evKeyDown Then
	Begin
		If KeyBufLen < 30 Then
		Begin
			KeyBuf[KeyBufLen] := Event.CharCode;
			Inc(KeyBufLen);
			If Event.CharCode = #0 Then
			Begin
				KeyBuf[KeyBufLen] := Chr(Event.ScanCode);
				Inc(KeyBufLen)
			End
		End
	End
End;

procedure TTerminal.StrWrite;
Label LF;
begin
	While Count > 0 Do
	Begin
		Case s^ Of
			#8:
				If Cursor.X>0 Then Dec(Cursor.X);
			#13:
				Cursor.X := 0;
			#10:
			Begin
LF:			Inc(Cursor.Y);
				While(Cursor.Y >= YRes)Do
				Begin
					Move(Buffer^[XRes], Buffer^, XRes*(YRes-1)*2);
					MoveChar(Buffer^[(YRes-1)*XRes], ' ', TextAttr, XRes);
					Dec(Cursor.Y);
				End
			End;
			Else
				If(Cursor.Y >= 0)And(Cursor.X >= 0)Then
					MoveChar(Buffer^[Cursor.Y*XRes + Cursor.X], s^, TextAttr, 1);
				Inc(Cursor.X);
				If Cursor.X >= XRes Then
				Begin
					Cursor.X := 0;
					Goto LF
				End;
		End;
		Inc(s);
		Dec(Count)
	End;

	Delta.X := Min(Cursor.X, Delta.X);
	Delta.X := Min(Cursor.X-Size.X+1, Delta.X+Size.X);
	Delta.X := Max(0, Delta.X);

	Delta.Y := Min(Cursor.Y, Delta.Y);
	Delta.Y := Min(Cursor.Y-Size.Y+1, Delta.Y+Size.Y);
	Delta.Y := Max(0, Delta.Y);

	ScrollTo(Delta.X, Delta.Y);
	DrawView
end;

{ Window Text Device Driver }
type
	WindowData = record
		Screen: PTextDevice;
		Filler: Array [1..12] of Char;
	end;

function WindowWrite(var F: TTextRec): Integer; far;
begin
  with F do
  begin
    WindowData(UserData).Screen^.StrWrite(BufPtr^, BufPos);
    BufPos := 0;
  end;
  WindowWrite := 0;
end;

function WindowRead(var F: TTextRec): Integer; far;
begin
  with F do
  begin
    BufPos := 0;
		BufEnd := WindowData(F.UserData).Screen^.StrRead(BufPtr^);
  end;
  WindowRead := 0;
end;

function WindowFlush(var F: TTextRec): Integer; far;
begin
	F.BufPos := 0;
	F.BufEnd := 0;
	WindowFlush := 0;
end;

function WindowOpen(var F: TTextRec): Integer; far;
begin
  with F do
  begin
    if Mode = fmInput then
    begin
      InOutFunc := @WindowRead;
      FlushFunc := @WindowFlush;
    end
    else
    begin
      InOutFunc := @WindowWrite;
      FlushFunc := @WindowWrite;
    end;
    WindowOpen := 0;
  end;
end;

function WindowIgnore(var F: TTextRec): Integer; far;
begin
  WindowIgnore := 0;
end;

procedure AssignDevice(var T: Text; Screen: PTextDevice);
begin
	with TTextRec(T) do
  begin
    Handle := $FFFF;
    Mode := fmClosed;
    BufSize := SizeOf(Buffer);
    BufPtr := @Buffer;
    OpenFunc := @WindowOpen;
    CloseFunc := @WindowIgnore;
    WindowData(UserData).Screen:= Screen;
	end
end;

Var OldExit: Procedure;
Procedure Quitti; Far;
Begin
	ExitProc := OldExit;
	Assign(Output, '');
	ReWrite(Output);
End;

Begin
	OldExit := ExitProc;
	ExitProc := Quitti;
end.
