; *******************************************************
; *                                                     *
; *     Turbo Pascal Run-time Library                   *
; *     Main module                                     *
; *                                                     *
; *******************************************************

        TITLE   MAIN

        INCLUDE SE.ASM

DATA    SEGMENT WORD PUBLIC
; Externals
        EXTRN   OvrHeapSize:WORD,OvrHeapOrg:WORD,OvrHeapPtr:WORD
        EXTRN   OvrHeapEnd:WORD,OvrLoadList:WORD,HeapOrg:DWORD
        EXTRN   HeapPtr:DWORD,HeapEnd:DWORD,FreeList:DWORD
        EXTRN   HeapError:DWORD,ExitProc:DWORD,ExitCode:WORD
        EXTRN   ErrorAddr:DWORD,PrefixSeg:WORD,InOutRes:WORD
        EXTRN   Test8086:BYTE
        extrn Terminator:dword
DATA    ENDS

CODE    SEGMENT BYTE PUBLIC

        ASSUME  CS:CODE,DS:DATA

; Externals
        EXTRN   ClearDSeg:NEAR

; Publics
        PUBLIC  InitTurbo,HaltTurbo,HaltError;,PrintString
        PUBLIC  Terminate,LibEntry,LibExit,LibExitProc,InitTask

        public ExecSwap

ExecSwap:
        arg FileName,dword,1
        arg Params,dword,1
        entry far

        exit

; Initialize runtime library. First instruction in any program
; is a call to this routine.
InitTurbo:
        MOV     DX,DATA                 ;Initialize DS
        MOV     DS,DX
        MOV     PrefixSeg,ES            ;Save PSP segment
        XOR     BP,BP                   ;End of stack frame chain             ; .ASM <- that is for BP
        CALL    ClearDSeg               ;Zero fill data segment
        CALL    Check8086               ;Determine CPU type
        MOV     AX,SP                   ;Compute first free segment
        ADD     AX,4+15                 ;address in AX
        MOV     CL,4
        SHR     AX,CL
        MOV     DX,SS
        ADD     AX,DX
        MOV     OvrHeapOrg,AX           ;Initialize overlay manager
        MOV     OvrHeapPtr,AX           ;variables
        ADD     AX,OvrHeapSize
        MOV     OvrHeapEnd,AX
        MOV     HeapOrg.seg,AX          ;Initialize heap manager
        MOV     HeapPtr.seg,AX          ;variables
        MOV     FreeList.seg,AX
        MOV     ES,PrefixSeg
        MOV     AX,ES:pspMemTop
        MOV     HeapEnd.seg,AX
        MOV     HeapError.ofs,OFFSET CS:HeapFailure
        MOV     HeapError.seg,CS

        RETF                            ;Back to main program

; CPU detection routine. Sets the Test8086 variable according to the
; type of CPU detected.
Check8086:
        XOR     AX,AX                   ;0 means 8088/8086
        PUSHF                           ;BX = Flags
        POP     BX
        AND     BH,0FH                  ;Clear bits 12-15
        PUSH    BX                      ;Flags = BX
        POPF
        PUSHF                           ;CX = Flags
        POP     CX
        AND     CH,0F0H                 ;Bits 12-15 set?
        CMP     CH,0F0H
        JE      @@1                     ;Yes, 8086
        INC     AX                      ;1 means 80286
        OR      BH,0F0H                 ;Set bits 12-15
        PUSH    BX                      ;Flags = BX
        POPF
        PUSHF                           ;CX = Flags
        POP     CX
        AND     CH,0F0H                 ;Bits 12-15 cleared?
        JE      @@1                     ;Yes, 80286
        INC     AX                      ;2 means 80386
@@1:    MOV     Test8086,AL             ;Save CPU test result
        RET

; Default heap error handler. Return 0 to indicate run-time error.
HeapFailure:
        XOR     AX,AX
        RETF    2

; Critical error interrupt handler. Control arrives here when
; DOS encounters a "critical error". Convert critical error code
; to Turbo Pascal I/O error code, and return.
Int24Handler:
        STI                             ;Enable interrupts
        ADD     SP,6                    ;Remove INT 24H return info
        POP     AX                      ;Get INT 21H AX register
        AND     DI,01FH                 ;Return AX=150..181
        ADD     DI,150
        CMP     AH,39H                  ;DOS 2.0 style function?
        JAE     @@1                     ;Yes, @@1
        MOV     DI,0FFFFH               ;Return AX=0FFFFH
@@1:    PUSH    DI                      ;Save error code
        MOV     AH,54H                  ;Dummy function call to get
        INT     DOS                     ;DOS into a stable state
        MOV     BP,SP                   ;Set CF in return flags
        OR      BYTE PTR [BP+22],1
        POP     AX                      ;Restore registers
        POP     BX
        POP     CX
        POP     DX
        POP     SI
        POP     DI
        POP     BP
        POP     DS
        POP     ES
        IRET                            ;Return to caller

; RunError standard procedure
HaltError:
        POP     CX
        POP     BX
        jmp short Terminate

; Dummy entry points
LibEntry:
LibExit:
LibExitProc:
InitTask:

; Halt standard procedure
HaltTurbo:
        XOR     CX,CX
        XOR     BX,BX

; Terminate program and return to DOS
; In    AX    = Exit code
;       BX:CX = Error address (or NIL)
Terminate:
        MOV DX,DATA                 ;Reset DS
        MOV DS,DX
        STI                         ;Enable interrupts
        MOV ExitCode,AX             ;Save exit code
        ;mov ErrAddr.ofs, cx
        ;mov ErrAddr.seg, bx
        jmp dword ptr Terminator

CODE    ENDS

        END
